from datetime import timedelta
from flask import Blueprint, request, url_for
from srht.api import get_results
from srht.config import cfg, get_origin
from srht.database import db
from srht.flask import csrf_bypass
from srht.oauth import AbstractOAuthService
from srht.oauth import DelegatedScope
from srht.cache import set_cache, expunge_cache
import json
import requests
import sys

scm_scopes = [
    DelegatedScope("info", "repository details", True),
    DelegatedScope("data", "version controlled data", False),
    DelegatedScope("access", "access control lists", True),
]

meta_origin = get_origin("meta.sr.ht")
builds_client_id = cfg("builds.sr.ht", "oauth-client-id", default=None)

class BaseScmOAuthService(AbstractOAuthService):
    def __init__(self, service_name, token_class, user_class, sshkey_class,
            service_origin=None, client_id=None, client_secret=None,
            required_scopes=None):
        client_id = client_id or cfg(service_name, "oauth-client-id")
        client_secret = client_secret or cfg(service_name, "oauth-client-secret")
        if not required_scopes:
            required_scopes = ["profile", "keys"] + ([
                    "{}/jobs:write".format(builds_client_id)
                ] if builds_client_id else [])

        super().__init__(client_id, client_secret,
                required_scopes=required_scopes,
                delegated_scopes=scm_scopes,
                token_class=token_class, user_class=user_class)
        self.service_name = service_name
        self.service_origin = service_origin or cfg(service_name, "origin")
        self.SSHKey = sshkey_class

    def cache_key(self, user, meta_key):
        b64key = meta_key["key"].split(" ")
        # The key should be of the form "ssh-<format> <hash> <user comment>"
        # We only need the key hash itself.
        if len(b64key) < 2:
            return False
        b64key = b64key[1]
        cache = {
            "user_id": user.id,
            "username": user.username,
        }
        set_cache(f"{self.service_name}.ssh-keys.{b64key}",
                timedelta(days=7), json.dumps(cache))

    def ensure_user_sshkey(self, user, meta_key):
        """
        Ensures this SSH key is registered with this user, and returns True if
        their authorized_keys file needs to be regenerated.

        `meta_key` should be the key object returned from meta.sr.ht.
        """
        key = self.SSHKey.query.filter(
                self.SSHKey.meta_id == meta_key["id"]).one_or_none()
        if key:
            self.cache_key(user, meta_key)
            return False
        key = self.SSHKey()
        key.user_id = user.id
        key.meta_id = meta_key["id"]
        key.key = meta_key["key"]
        key.fingerprint = meta_key["fingerprint"]
        db.session.add(key)
        db.session.commit()
        self.cache_key(user, meta_key)
        return True

    def ensure_meta_webhooks(self, user, webhooks):
        origin = self.service_origin
        webhook_url = origin + url_for("webhooks_notify.notify_keys")
        webhooks.update({
            webhook_url: ["ssh-key:add", "ssh-key:remove"]
        })
        return super().ensure_meta_webhooks(user, webhooks)

    def lookup_or_register(self, token, token_expires, scopes):
        user = super().lookup_or_register(token, token_expires, scopes)
        db.session.flush()
        keys_url = f"{meta_origin}/api/user/ssh-keys"
        for key in get_results(keys_url, user.oauth_token):
            self.ensure_user_sshkey(user, key)
        db.session.commit()
        return user

def make_webhooks_notify_blueprint(module_name, oauth_service,
        blueprint_name="webhooks_notify"):
    """
    Creates a flask blueprint that implements the boilerplate code for
    add/removing sshkeys in the given service's tables.
    """
    webhooks_notify = Blueprint(blueprint_name, module_name)

    service_name = oauth_service.service_name
    User = oauth_service.User
    SSHKey = oauth_service.SSHKey

    @csrf_bypass
    @webhooks_notify.route("/webhook/notify/keys", methods=["POST"])
    def notify_keys():
        payload = json.loads(request.data.decode('utf-8'))
        event = request.headers.get("X-Webhook-Event")
        if event == "ssh-key:add":
            user = User.query.filter(
                    User.username == payload["owner"]["name"]).one_or_none()
            oauth_service.ensure_user_sshkey(user, payload)
            return "Added user's SSH key, thanks!"
        elif event == "ssh-key:remove":
            key = SSHKey.query.filter(
                    SSHKey.meta_id == payload["id"]).one_or_none()
            if not key:
                return "Wasn't familiar with this key anyway."
            b64key = key.key.split(" ")
            if len(b64key) >= 2:
                b64key = b64key[1]
                expunge_cache(f"{service_name}.ssh-keys.{b64key}")
            if key:
                db.session.delete(key)
                db.session.commit()
            return "Removed user's SSH key, thanks!"
        return f"Unexpected event {event}"

    return webhooks_notify
